// The MIT License

// Copyright (c) 2010 Massachusetts Institute of Technology

// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:

// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.

// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

// Author: Richard Uhler ruhler@mit.edu

import ClientServer::*;
import FIFO::*;
import FixedPoint::*;
import GetPut::*;

typedef Server#(
    word,
    Tuple2#(word, word)
) SquareRoot#(type word);

// Square root
// Input: x
// Output: q, r
//  where x = q*q + r.
// 
// Algorithm developed based on description at
// http://www.itl.nist.gov/div897/sqg/dads/HTML/squareRoot.html
//
// itersPerCycle specifies how many iterations of the algorithm to perform
// each clock cycle. This should divide evenly half the bitwidth of the
// operand.
module mkSquareRoot(Integer itersPerCycle, SquareRoot#(Bit#(ws)) ifc);

    if (valueof(ws) % (2 * itersPerCycle) != 0) begin
        error("itersPerCycle must evenly divide half the bitwidth");
    end

    FIFO#(Bit#(ws)) incoming <- mkFIFO();
    FIFO#(Tuple2#(Bit#(ws), Bit#(ws))) outgoing <- mkFIFO();

    Reg#(Bool) busy <- mkReg(False);
    Reg#(Bit#(ws)) qReg <- mkRegU();
    Reg#(Bit#(ws)) rReg <- mkRegU();
    Reg#(Bit#(ws)) xReg <- mkRegU();
    Reg#(Bit#(TLog#(ws))) iReg <- mkRegU();

    rule start (!busy);
        let xi <- toGet(incoming).get();

        busy <= True;
        xReg <= xi;
        rReg <= 0;
        qReg <= 0;
        iReg <= 0;
    endrule

    // Get the top 2 bits of a bit vector
    function Bit#(2) top2(Bit#(n) a) = a[valueof(n)-1:valueof(n)-2];

    // Perform a single iteration of the square root algorithm.
    // Returns new r,q,x
    function Tuple3#(Bit#(ws), Bit#(ws), Bit#(ws)) iterate(Bit#(ws) r, Bit#(ws) q, Bit#(ws) x);
        Bit#(TAdd#(2, ws)) d = {q, 2'b01};
        Bit#(TAdd#(2, ws)) r2 = {r, top2(x)};

        if (d > r2) begin
            q = q << 1;
            r = r2[valueof(ws)-1:0];
        end else begin
            q = (q << 1) | 1;
            Bit#(TAdd#(2, ws)) diff = r2 - d;
            r = diff[valueof(ws)-1:0];
        end
        
        return tuple3(r, q, x << 2);

    endfunction

    rule doiterate (busy);
        Bit#(ws) r = rReg;
        Bit#(ws) q = qReg;
        Bit#(ws) x = xReg;

        for (Integer i = 0; i < itersPerCycle; i = i+1) begin
            let iout = iterate(r, q, x);
            r = tpl_1(iout);
            q = tpl_2(iout);
            x = tpl_3(iout);
        end

        if (iReg + fromInteger(itersPerCycle) == fromInteger(valueof(ws)/2)) begin
            outgoing.enq(tuple2(q, r));
            busy <= False;
        end else begin
            iReg <= iReg + fromInteger(itersPerCycle);
        end

        rReg <= r;
        qReg <= q;
        xReg <= x;
    endrule

    interface Put request = toPut(incoming);
    interface Get response = toGet(outgoing);

endmodule

// Fixed Point square root
module mkFixedPointSquareRoot(Integer itersPerCycle, SquareRoot#(FixedPoint#(iw, fw)) ifc)
    provisos(
        Add#(a__, TAdd#(iw, fw), TAdd#(iw, TMul#(2, fw)))
    );

    // We use the integer square root algorithm to do fixed point square root.
    // 
    // If you pack a fixed point number x,
    // you get a new number which is x * 2^fw
    // If you unpack a number y into a fixed point number, you get a fixed point
    // number which is y / 2^fw.
    //
    // Let the input be a fixed point number x. We want to generate the fixed
    // point number sqrt(x).
    // We unpack x, multiply by 2^fw, perform integer square root, and get the
    // integer:
    // sqrt((x * 2^fw) * 2^fw) = sqrt(x) * 2^fw
    // Simply unpack and we get our fixed point sqrt(x).

    SquareRoot#(Bit#(TAdd#(iw, TMul#(2, fw)))) sqrt <- mkSquareRoot(itersPerCycle);

    interface Put request;
        method Action put(FixedPoint#(iw, fw) x);
            Bit#(TAdd#(iw, fw)) xx = pack(x);
            Bit#(TAdd#(iw, TMul#(2, fw))) xxx = zeroExtend(xx);
            xxx = xxx << valueof(fw);

            sqrt.request.put(xxx);
        endmethod
    endinterface

    interface Get response;
        method ActionValue#(Tuple2#(FixedPoint#(iw, fw), FixedPoint#(iw, fw))) get();
            match {.q, .r} <- sqrt.response.get();
            Bit#(TAdd#(iw, fw)) qb = q[valueof(iw) + valueof(fw)-1:0];
            FixedPoint#(iw, fw) qf = unpack(qb);

            // TODO: return a useful remainder
            return tuple2(qf, ?);
        endmethod
    endinterface
endmodule
            
